/// Functions for performing cryptographic operations on the main data structures.

use super::*;
use openssl::{
    error::ErrorStack,
    encrypt::{Encrypter, Decrypter},
    pkey::{PKey, Public, Private},
    symm::{Cipher, encrypt as openssl_encrypt, decrypt as openssl_decrypt},
    nid::Nid,
    ec::{EcGroup, EcKey},
    rand::rand_bytes,
};
use serde_block_tree::{self, to_vec, from_vec};
use serde::de::{DeserializeOwned};

/// Data that may or may not be encrypted.
#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub enum Cryptotext<T: Serialize> {
    /// The inner value of `T` is plaintext.
    Plain(T),
    /// The inner value is ciphertext.
    Cipher(Vec<u8>),
}

/// Errors that can occur during cryptographic operations.
#[derive(Debug)]
pub enum Error {
    NoReadCap,
    NoKeyAvailable,
    MissingPrivateKey,
    Message(String),
    Serde(serde_block_tree::Error),
}

impl From<ErrorStack> for Error {
    fn from(error: ErrorStack) -> Error {
        Error::Message(error.to_string())
    }
}

impl From<serde_block_tree::Error> for Error {
    fn from(error: serde_block_tree::Error) -> Error {
        Error::Serde(error)
    }
}

pub type Result<T> = std::result::Result<T, Error>;

/// A cryptographic hash.
#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Hashable, Clone)]
pub enum Hash {
    Sha2_256([u8; 32]),
    #[serde(with = "BigArray")]
    Sha2_512([u8; 64]),
}

/// A cryptographic signature.
#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub enum Signature {
    #[serde(with = "BigArray")]
    Secp256k1([u8; 64]),
}

/// Identifies a type of cryptographic key. The variants of this enum match those of `Key`.
pub enum KeyId {
    // TODO: Write a macro to generate this from `Key`.
    Aes256Cbc,
    Aes256Ctr,
    Secp256k1,
}

/// A cryptographic key.
#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub enum Key {
    /// A key for the AES 256 cipher in Cipher Block Chaining mode. Note that this includes the
    /// initialization vector, so that a value of this variant contains all the information needed
    /// to fully initialize a cipher context.
    Aes256Cbc {
        key: [u8; 32],
        iv: [u8; 16],
    },
    /// A key for the AES 256 cipher in counter mode.
    Aes256Ctr {
        key: [u8; 32],
        iv: [u8; 16]
    },
    Secp256k1 {
        #[serde(with = "BigArray")]
        public: [u8; 88],
        private: Option<Vec<u8>>
    },
}

impl Key {
    /// Returns the key part of this `Key`. Note that if this `Key` is for an asymmetric algorithm,
    /// then this returns the *private* key (if it is present).
    fn key_slice(&self) -> Option<&[u8]> {
        match self {
            Key::Aes256Cbc { key, .. } => Some(key),
            Key::Aes256Ctr { key, .. } => Some(key),
            Key::Secp256k1 { private, .. } => match private {
                Some(key) => Some(key.as_slice()),
                None => None,
            },
        }
    }

    /// Returns the IV part of this`Key`.
    fn iv_slice(&self) -> Option<&[u8]> {
        match self {
            Key::Aes256Cbc { iv, .. } => Some(iv),
            _ => None,
        }
    }

    fn generate(id: KeyId) -> Result<Key> {
        match id {
            KeyId::Aes256Cbc => {
                let mut key = [0; 32];
                let mut iv = [0; 16];
                rand_bytes(&mut key).map_err(Error::from)?;
                rand_bytes(&mut iv).map_err(Error::from)?;
                Ok(Key::Aes256Cbc { key, iv })
            },
            KeyId::Aes256Ctr => {
                let mut key = [0; 32];
                let mut iv = [0; 16];
                rand_bytes(&mut key).map_err(Error::from)?;
                rand_bytes(&mut iv).map_err(Error::from)?;
                Ok(Key::Aes256Ctr { key, iv })
            },
            KeyId::Secp256k1 => {
                let group = EcGroup::from_curve_name(Nid::SECP256K1).map_err(Error::from)?;
                let key = EcKey::generate(&group).map_err(Error::from)?;
                let mut public = [0; 88];
                let public_vec = key.public_key_to_der().map_err(Error::from)?;
                public.copy_from_slice(&public_vec);
                let private = Some(key.private_key_to_der().map_err(Error::from)?);
                Ok(Key::Secp256k1 { public, private })
            },
        }
    }
}

enum EncryptionAlgo<'a> {
    Symmetric {
        cipher: Cipher,
        key: &'a [u8],
        iv: Option<&'a [u8]>
    },
    Asymmetric(PKey<Public>),
}

impl<'a> EncryptionAlgo<'a> {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        match self {
            EncryptionAlgo::Symmetric { cipher, key, iv } => {
                openssl_encrypt(*cipher, key, *iv, slice).map_err(Error::from)
            },
            EncryptionAlgo::Asymmetric(pkey) => {
                let encrypter = Encrypter::new(pkey).map_err(Error::from)?;
                let buffer_len = encrypter.encrypt_len(slice).map_err(Error::from)?;
                let mut ciphertext = vec![0; buffer_len];
                let ciphertext_len = encrypter.encrypt(slice, &mut ciphertext)
                    .map_err(Error::from)?;
                ciphertext.truncate(ciphertext_len);
                Ok(ciphertext)
            }
        }
    }
}

impl<'a> TryFrom<&'a Key> for EncryptionAlgo<'a> {
    type Error = Error;
    fn try_from(key: &'a Key) -> Result<EncryptionAlgo<'a>> {
        match key {
            Key::Aes256Cbc { key: key_slice, iv } => Ok(EncryptionAlgo::Symmetric {
                cipher: Cipher::aes_256_cbc(),
                key: key_slice,
                iv: Some(iv),
            }),
            Key::Aes256Ctr { key: key_slice, iv } => Ok(EncryptionAlgo::Symmetric {
                cipher: Cipher::aes_256_ctr(),
                key: key_slice,
                iv: Some(iv),
            }),
            Key::Secp256k1 { public, .. } => {
                let pkey = PKey::public_key_from_der(public.as_slice())
                    .map_err(|err| Error::Message(err.to_string()));
                Ok(EncryptionAlgo::Asymmetric(pkey?))
            },
        }
    }
}

enum DecryptionAlgo<'a> {
    Symmetric {
        cipher: Cipher,
        key: &'a [u8],
        iv: Option<&'a [u8]>
    },
    Asymmetric(PKey<Private>),
}

impl<'a> DecryptionAlgo<'a> {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        match self {
            DecryptionAlgo::Symmetric { cipher, key, iv } => {
                openssl_decrypt(*cipher, key, *iv, slice).map_err(Error::from)
            },
            DecryptionAlgo::Asymmetric(pkey) => {
                let decrypter = Decrypter::new(pkey).map_err(Error::from)?;
                let buffer_len = decrypter.decrypt_len(slice).map_err(Error::from)?;
                let mut plaintext = vec![0; buffer_len];
                let plaintext_len = decrypter.decrypt(slice, &mut plaintext)
                    .map_err(Error::from)?;
                plaintext.truncate(plaintext_len);
                Ok(plaintext)
            },
        }
    }
}

impl<'a> TryFrom<&'a Key> for DecryptionAlgo<'a> {
    type Error = Error;
    fn try_from(key: &'a Key) -> Result<DecryptionAlgo<'a>> {
        match key {
            Key::Aes256Cbc { key: key_slice, iv } => Ok(DecryptionAlgo::Symmetric {
                cipher: Cipher::aes_256_cbc(),
                key: key_slice,
                iv: Some(iv),
            }),
            Key::Aes256Ctr { key: key_slice, iv } => Ok(DecryptionAlgo::Symmetric {
                cipher: Cipher::aes_256_ctr(),
                key: key_slice,
                iv: Some(iv),
            }),
            Key::Secp256k1 { private, .. } => {
                let private = private.as_ref().ok_or(Error::MissingPrivateKey)?;
                let pkey = PKey::private_key_from_der(private.as_slice()).map_err(Error::from);
                Ok(DecryptionAlgo::Asymmetric(pkey?))
            }
        }
    }
}

pub(crate) fn encrypt_block(
    versioned_block: VersionedBlock, principal: &Principal, key: &Key
) -> Result<VersionedBlock> {
    let VersionedBlock::V0(mut block) = versioned_block;
    let body = match block.body {
        Cryptotext::Cipher(_) => return Ok(VersionedBlock::V0(block)),
        Cryptotext::Plain(body) => body
    };
    let (principal_owned, read_cap) = block.read_caps.remove_entry(principal)
        .ok_or(Error::NoReadCap)?;
    let block_key = decrypt(read_cap, key)?;
    let new_body = encrypt_slice(&body, &block_key)?;
    block.body = Cryptotext::Cipher(new_body);
    block.read_caps.insert(principal_owned, encrypt(&block_key, key)?);
    Ok(VersionedBlock::V0(block))
}

pub(crate) fn decrypt_block(
    versioned_block: VersionedBlock, principal: &Principal, key: &Key
) -> Result<VersionedBlock> {
    let VersionedBlock::V0(mut block) = versioned_block;
    let body = match block.body {
        Cryptotext::Plain(_) => return Ok(VersionedBlock::V0(block)),
        Cryptotext::Cipher(body) => body
    };
    let (principal_owned, read_cap) = block.read_caps.remove_entry(principal)
        .ok_or(Error::NoReadCap)?;
    let block_key = decrypt(read_cap, key)?;
    let new_body = decrypt_slice(&body, &block_key)?;
    block.body = Cryptotext::Plain(new_body);
    block.read_caps.insert(principal_owned, Cryptotext::Plain(block_key));
    Ok(VersionedBlock::V0(block))
}

fn encrypt<T: Serialize>(value: &T, key: &Key) -> Result<Cryptotext<T>> {
    let data = to_vec(value).map_err(Error::from)?;
    let vec = encrypt_slice(&data, key);
    Ok(Cryptotext::Cipher(vec?))
}

fn decrypt<T: Serialize + DeserializeOwned>(cryptotext: Cryptotext<T>, key: &Key) -> Result<T> {
    let data = match cryptotext {
        Cryptotext::Plain(value) => return Ok(value),
        Cryptotext::Cipher(data) => data
    };
    let vec = decrypt_slice(&data, key);
    from_vec(&vec?).map_err(Error::from)
}

fn encrypt_slice(plaintext: &[u8], key: &Key) -> Result<Vec<u8>> {
   let algo = EncryptionAlgo::try_from(key)?;
   algo.encrypt(plaintext)
}

fn decrypt_slice(ciphertext: &[u8], key: &Key) -> Result<Vec<u8>> {
    let algo = DecryptionAlgo::try_from(key)?;
    algo.decrypt(ciphertext)
}

#[cfg(test)]
mod tests {
    use super::*;
    use test_helpers::*;

    #[test]
    fn key_aes256cbc_key_slice_is_correct() -> Result<()> {
        let key = Key::Aes256Cbc { key: KEY, iv: IV };
        let key_part = key.key_slice().ok_or(Error::NoKeyAvailable);
        assert_eq!(KEY, key_part?);
        Ok(())
    }

    #[test]
    fn key_aes256cbc_iv_slice_is_correct() {
        let key = Key::Aes256Cbc { key: KEY, iv: IV };
        let iv_part = key.iv_slice().unwrap();
        assert_eq!(IV, iv_part);
    }

    fn encrypt_decrypt_block_test_case(
        mut actual: VersionedBlock, principal: &Principal, key: &Key
    ) -> Result<()> {
        let expected = actual.clone();
        actual = encrypt_block(actual, principal, key)?;
        actual = decrypt_block(actual, principal, key)?;
        assert_eq!(expected, actual);
        Ok(())
    }

    fn make_versioned_block(principal: Principal, block_key: Key) -> Result<VersionedBlock> {
        let mut block = make_block()?;
        block.read_caps.clear();
        block.read_caps.insert(principal, Cryptotext::Plain(block_key));
        Ok(VersionedBlock::V0(block))
    }

    #[test]
    fn encrypt_decrypt_block_aes256cbc() -> Result<()> {
        let principal = make_principal();
        let block_key = Key::Aes256Cbc { key: BLOCK_KEY, iv: BLOCK_IV };
        let block = make_versioned_block(principal.clone(), block_key)?;
        let key = Key::Aes256Cbc { key: KEY, iv: IV };
        encrypt_decrypt_block_test_case(block, &principal, &key)
    }

    #[test]
    fn encrypt_decrypt_block_aes256ctr() -> Result<()> {
        let principal = make_principal();
        let block_key = Key::Aes256Ctr { key: BLOCK_KEY, iv: BLOCK_IV };
        let block = make_versioned_block(principal.clone(), block_key)?;
        let key = Key::Aes256Ctr { key: KEY, iv: IV };
        encrypt_decrypt_block_test_case(block, &principal, &key)
    }

    /// Tests that validate the dependencies of this module.
    mod dependency_tests {
        use super::*;

        /// This test validates that data decrypted with the `Crypter` API matches data that was
        /// previously encrypted using it.
        #[test]
        fn crypter() {
            use super::*;
            let expected = b"We attack at the crack of noon!";
            let cipher = Cipher::aes_256_cbc();

            let ciphertext = openssl_encrypt(cipher, &KEY, Some(&IV), expected).unwrap();
            let actual = openssl_decrypt(cipher, &KEY, Some(&IV), ciphertext.as_slice()).unwrap();

            assert_eq!(expected, actual.as_slice());
        }

        /// Tests that the keys for the SECP256K1 curve are the expected sizes.
        #[test]
        fn secp256k1_key_lengths() {
            let group = EcGroup::from_curve_name(Nid::SECP256K1).unwrap();
            let key = EcKey::generate(&group).unwrap();
            let public = key.public_key_to_der().unwrap();
            let private = key.private_key_to_der().unwrap();
            let public_len = public.len();
            let private_len = private.len();
            assert_eq!(88, public_len);
            assert_eq!(118, private_len);
        }
    }
}
